#!/usr/bin/env perl

#
#   Copyright (C) 2010 by G. Montagner <gmontagner@sorint.it> and others
#
#   This script is free software; you can redistribute it and/or modify 
#   it under the same terms as Perl itself, either Perl version 5.8.8 or, 
#   at your option, any later version of Perl 5 you may have available.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Credits: 
#   Connection routines taken from Jason Lancaster's "Send NSCA Perl Client",
#   data routines and usage ported (by Jason Lancaster) from Ethan Galstad's 
#   NSCA package

my $info = <<INFO;

NPRD - Nagios Perl Replicator Daemon - v1.0
Last modified 2010/12/27 by Giacomo Montagner <gmontagner AT sorint.it>
Connection routines taken from Jason Lancaster's "Send NSCA Perl Client", 
data routines and usage ported from Ethan Galstad's NSCA package.

Usage: $0 -c <configuration file> [ -f ] [ -h|--help ]

Options: 
-f 	Foreground - do not daemonize
-h	Print this help message
--help	Same as above

Note:
The only supported data encryption method is XOR.

INFO

use strict;
use warnings;
use Getopt::Long;
use Socket; 
use Daemon::Daemonize qw /daemonize/;
use Sys::Syslog;
use DBI;
use Fcntl qw(:flock);

# Main program's ugly global variables 
my $head = "Nagios Perl Replicator Daemon - Ver 1.0";
my $foreground = 0;	# Default to daemon
my (@crc32_table);    # we really don't need this to be global
our $conffile;
our %conf;
my $help;

# DB queries
my $hostq = 'select nh.display_name as host,return_code,output 
	from nagios_hostchecks nhc, nagios_hosts nh 
	where nh.host_object_id=nhc.host_object_id';
my $serviceq = 'select nh.display_name as host, ns.display_name as service, return_code, output 
	from nagios_hosts nh, nagios_services ns, nagios_servicechecks nsc 
	where nh.host_object_id = ns.host_object_id and nsc.service_object_id = ns.service_object_id';

# Syslog facility:
my $ident = 'NPRD';
my $logopts = 'ndelay,pid,perror';
my $facility = 'LOG_DAEMON';

# Define options
Getopt::Long::Configure('bundling');
GetOptions("c=s" => \$conffile, "f" => \$foreground, "h|help" => \$help);

# Check if config file was specified or if help was requested
if (defined($help) || !(defined($conffile))) {print "$info"; exit 1; } ;

#######################
#     subroutines     # 
#######################

sub syslog_die {
	syslog "LOG_ERR", "@_";
	exit 1;
}

sub readconf {
	my ($conffile) = @_;
	my %configuration;
	open(CONF, '<', $conffile) or die "Cannot read configuration file $conffile";
	my @conf = <CONF>;
	close(CONF);
	foreach (@conf) {
		chomp;
		next if (/^\s*\#.*$|^\s*$/);	# Comment or empty line
		s/\#.*//;	# Remove inline comments
		s/['"]//g;
		my ($key, $val) = (/^\s*(.*?)\s*=\s*(.*?)\s*$/);
		$configuration{$key} = $val;
	}
	return %configuration;
}

sub printconf {
	syslog "LOG_INFO", "Active configuration:";
	my (%configuration) = @_;
	while (my ($key, $val) = each(%configuration)) {
		syslog "LOG_INFO", "    $key = $val" unless ($key =~ /passwd/);
	}
	syslog "LOG_INFO", "    dbpasswd was (intentionally) not printed.";
	return 1;
}

sub writepid {
	my ($pidfile) = @_;
	if (open(my $pfh, '>', "$pidfile")) {
		my $previous_default = select($pfh);
		$|++;
		select($previous_default);
		print $pfh "$$";
		return $pfh;
	} else {
		syslog "LOG_WARNING", "Cannot write to pidfile '$pidfile' - check your privileges.";
		return ;
	}
}

sub get_timestamp {
	my ($delay) = @_;
	my $time = time();
	$time += $delay if defined($delay);
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($time);
	return sprintf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d", ($year + 1900), ($mon + 1), $mday, $hour, $min, $sec);
}

# xor the data (the only type of "encryption" we currently use)
sub myxor {
	my ($xor_key, $str) = @_;
 
	my $xor_str = $str ^ ($xor_key x int((length($str) + 127) / 128));
 
	return substr($xor_str, 0, length($str));
}

# build the crc table - must be called before calculating the crc value 
sub generate_crc32_table {
	my ($crc, $poly, $i, $j);

	$poly = 0xEDB88320;
	for($i = 0; $i < 256; $i++) {
		$crc = $i;
		for($j = 8; $j > 0; $j--) {
			if ($crc & 1) {
				$crc = ($crc >> 1) ^ $poly;
			} else {
				$crc >>= 1;
			}
		}
		$crc32_table[$i] = $crc;
	}

	return;
}

# calculates the CRC 32 value for a buffer
sub calculate_crc32 {
	my ($buffer) = @_;
	my ($crc, $this_char, $current_index);

	$crc = 0xFFFFFFFF;

	for($current_index = 0; $current_index < length($buffer); $current_index++) {
		$this_char = ord(substr($buffer, $current_index, 1));
		$crc = (($crc >> 8) & 0x00FFFFFF) ^ $crc32_table[($crc ^ $this_char) & 0xFF];
	}

	return ($crc ^ 0xFFFFFFFF);
}


sub build_packet {
	my ($timestamp, $xor_key, $hostname, $output, $return_code, $service) = @_;
	my $packet_version = "3";
	my $crc = 0;
	$output = "NO OUTPUT" unless defined($output);

 	# Build our packet.
	my $tobecrced;
	if (defined($service)) {
	 	$tobecrced = pack("nxx N a4 n a64 a128 a512xx",
 			$packet_version, $crc, $timestamp, $return_code, $hostname, $service, $output);
	} else {
	 	$tobecrced = pack("nxx N a4 n a64 a128 a512xx",
 			$packet_version, $crc, $timestamp, $return_code, $hostname, '', $output);
	}
 
 	# Get a signature for the packet.
 	$crc = calculate_crc32($tobecrced);
 
 	# Build the final packet with the sig.
	my $str;
	if (defined($service)) {
	 	$str = pack("nxx N a4 n a64 a128 a512xx",
 			$packet_version, $crc, $timestamp, $return_code, $hostname, $service, $output);
	} else {
	 	$str = pack("nxx N a4 n a64 a128 a512xx",
 			$packet_version, $crc, $timestamp, $return_code, $hostname, '', $output);
	}
 
 	# Xor the sucker.
 	my $string_to_send = myxor($xor_key, $str);

	return $string_to_send
}

sub run_query {
	# Start the timer. We don't want stale processes!
	local $SIG{ALRM} = sub { syslog "LOG_WARNING", "Query took too long to execute"; return; };       
	alarm 15;
	my ($dbh, $query) = @_;
	my %hash;
	my $sth = $dbh->prepare($query);
	if (!$sth) {
		syslog "LOG_WARNING", "Error:" . $dbh->errstr;
		return;
	}
	if (!$sth->execute) {
		syslog "LOG_WARNING", "Error:" . $sth->errstr;
		return;
	}
	while (my $row = $sth->fetchrow_hashref()) {
		my $host = $row->{'host'};
		if (defined($row->{'service'})) {
			my $service = $row->{'service'}; 
			print "Found service check for: $host, $service\n";
			$hash{$host}{$service}{'output'} = $row->{'output'};
			print "Plugin output: ".$hash{$host}{$service}{'output'}."\n";
			$hash{$host}{$service}{'return_code'} = $row->{'return_code'};
			print "Plugin return code: ".$hash{$host}{$service}{'return_code'}."\n";
		} else {
			print "Found host check for: $host\n";
			$hash{$host}{'output'} = $row->{'output'};
			print "Plugin output: ".$hash{$host}{'output'}."\n";
			$hash{$host}{'return_code'} = $row->{'return_code'};
			print "Plugin return code: ".$hash{$host}{'return_code'}."\n";
		}
	}
	alarm 0;
	return %hash;
}


sub connect_to {
	my ($paddr, $proto) = @_;
	my $failure = sprintf("Could not establish a connection to %s! Are you an allowed_host?\n\n", $conf{'remote_host'});
	my ($sock, $timestamp, $xor_key);
	if (! socket ($sock, PF_INET, SOCK_STREAM, $proto)) {
		syslog "LOG_WARNING", $failure;
		return;
	}
	if (! connect($sock, $paddr)) {
		syslog "LOG_WARNING", $failure;
		return;
	}

	# Get 128bit xor key and 4bit timestamp.
	if (! read($sock,$xor_key,128)) {
		syslog "LOG_WARNING", $failure;
		return;
	}
	if (! read($sock,$timestamp, 4)) {
		syslog "LOG_WARNING", $failure;
		return;
	}
	return ($sock, $timestamp, $xor_key);
}

sub logopen {
	my ($logfile) = @_;
	my $fh;
	if (! -f $logfile) {
		open (LOG, '>>', $logfile) or syslog_die "Unable to create file $logfile";
		close (LOG);
	}

	open($fh, '+<', $logfile) or syslog_die "Unable to open $logfile";
	flock($fh, LOCK_EX|LOCK_NB) or syslog_die "Unable to acquire lock on $logfile; is another daemon already running?";
	my $previous_default = select($fh);
	$|++;
	select($previous_default);
	return $fh;
}

###########
# SIGNALS #
###########

$SIG{USR1} = sub { printconf(%conf); };
$SIG{INT} = sub { syslog "LOG_INFO", "Exiting"; exit 0; };
$SIG{TERM} = sub { syslog "LOG_INFO", "Exiting"; exit 0; };


########
# MAIN # 
########

print "$head starting...\n";

# Read configuration
%conf = readconf($conffile);
undef $conffile;

# Init syslog facility:
openlog $ident, $logopts, $facility or die "Cannot log to syslog";

# Become daemon
if ($foreground != 1) {
	daemonize(chdir => '/', close => 'std') || die "Cannot become daemon";
	syslog "LOG_INFO", "$head :: Daemon Started";
}
printconf(%conf);

my $lockfh = logopen($conf{'log'});
# Print pid to pidfile:
my $pidfh = writepid($conf{'pid_file'}) if (defined($conf{'pid_file'}));

# Init database connection
my $dsn = sprintf("DBI:mysql:database=%s;host=%s;port=%s", $conf{'dbname'}, $conf{'dbhost'}, $conf{'dbport'});
my $dbh = DBI->connect($dsn, $conf{'dbuser'}, $conf{'dbpasswd'});
if (!(defined($dbh))) {
	syslog 'LOG_ERR', sprintf("Cannot connect to MySQL database %s on %s:%s as %s", $conf{'dbname'}, $conf{'dbhost'}, $conf{'dbport'}, $conf{'dbuser'});
	exit 1;
};
syslog "LOG_INFO", "MySQL connection: connected to $conf{'dbhost'}:$conf{'dbport'}";
my $drh = DBI->install_driver("mysql");

# Prepare connection to remote NSCA
my ($iaddr, $paddr, $proto, $xor_key, $timestamp);
my ($d);

$iaddr   = inet_aton($conf{'remote_host'});
$paddr   = sockaddr_in($conf{'port'}, $iaddr);
$proto   = getprotobyname('tcp');

# Generate the crc table. This will be used to sign the packet.
# We ported the code for this from the nsca_send.c file
generate_crc32_table();

# Enter main loop: 
my $exit = 0;
my $last_timestamp;
while ($exit == 0) {
	$last_timestamp = get_timestamp(-$conf{'interval'}) unless defined($last_timestamp);
	my $now = get_timestamp();

	seek($lockfh, 0, 0);
	print $lockfh "$now";
	
	# Get latest checks from the DB
	my $hostquery = $hostq." and nhc.start_time > '".$last_timestamp."' order by nhc.start_time;";
	my $servicequery = $serviceq." and nsc.start_time > '".$last_timestamp."' order by nsc.start_time;";
	my %host_checks = run_query($dbh, $hostquery);
	my %service_checks = run_query($dbh, $servicequery);

	if (%host_checks || %service_checks) {

		my @checks; 
		my ($sock, $timestamp, $xor_key) = connect_to($paddr, $proto);
	
		if ((!defined($sock)) || (!defined($timestamp)) || (!defined($xor_key))) {
			syslog "LOG_WARNING", sprintf("Failed to connect to NSCA %s:%s", $conf{'remote_host'}, $conf{'port'});
			sleep $conf{'interval'};
			next;
		}
		syslog "LOG_DEBUG", sprintf("NSCA connection: connected to %s:%s", $conf{'remote_host'}, $conf{'port'});
		
		foreach my $host (keys(%host_checks)) {
			my $str = build_packet($timestamp, $xor_key, $host, $host_checks{$host}{'output'}, $host_checks{$host}{'return_code'}, '');
			push(@checks, $str);
		}
		undef %host_checks;
		foreach my $host (keys(%service_checks)) {
			foreach my $service (keys(%{$service_checks{$host}})) {
				my $str = build_packet($timestamp, $xor_key, $host, $service_checks{$host}{$service}{'output'}, $service_checks{$host}{$service}{'return_code'}, $service);
				push(@checks, $str);
			}
		}
		undef %service_checks;
	
		if ($#checks > -1) {
			if ($#checks > $conf{'external_command_buffer_slots'}) {
				syslog "LOG_WARNING", "The number of checks is greater than external_command_buffer_slots size, some checks could not be sent";
				my $first = $#checks - $conf{'external_command_buffer_slots'} + 1;
				my $last = $#checks + 1;
				@checks = splice(@checks, $first, $last);
			}
			my $sent = 0;
			my $count = 0;
		
			foreach my $check (@checks) {
				$count++;
				if (send($sock,$check,0)) {
					$sent++;
				} else {
					syslog "LOG_WARNING", sprintf("Could not send packet %d\n", $count);
				}
			}
		}
		close($sock);
		syslog "LOG_DEBUG", sprintf("Closed NSCA connection to  %s:%s", $conf{'remote_host'}, $conf{'port'});
	} else {
		syslog "LOG_DEBUG", "No checks to submit, waiting ".$conf{'interval'}."s and retrying";
	}
	$last_timestamp = $now;
	sleep $conf{'interval'};
}

flock ($lockfh, LOCK_UN);
close ($lockfh);
close ($pidfh) if (defined($pidfh));

# Goodbye
exit 0;

